@@ -51,10 +51,11 @@ typedef struct php_rar_stream_data_t {
 	struct RAROpenArchiveDataEx	open_data;
 	struct RARHeaderDataEx		header_data;
 	HANDLE						rar_handle;
+	size_t						file_size;
 	/* TODO: consider encapsulating a php memory/tmpfile stream */
 	unsigned char				*buffer;
 	size_t						buffer_size;
-	size_t						buffer_cont_size; /* content size */
+	size_t						buffer_read_size; /* content size */
 	size_t						buffer_pos;
 	uint64						cursor;
 	int							no_more_data;
@@ -98,7 +99,11 @@ static char *_rar_wide_to_utf_with_alloc(const wchar_t *wide, int len)
 /* {{{ RAR file streams */
 
 /* {{{ php_rar_ops_read */
+#if PHP_VERSION_ID < 70400
 static size_t php_rar_ops_read(php_stream *stream, char *buf, size_t count TSRMLS_DC)
+#else
+static ssize_t php_rar_ops_read(php_stream *stream, char *buf, size_t count)
+#endif
 {
 	size_t n = 0;
 	STREAM_DATA_FROM_STREAM
@@ -113,12 +118,12 @@ static size_t php_rar_ops_read(php_stream *stream, char *buf, size_t count TSRML
 		while (left > 0) {
 			size_t this_read_size;
 			/* if nothing in the buffer or buffer already read, fill buffer */
-			if (/*self->buffer_cont_size == 0 || > condition not necessary */
-				self->buffer_pos == self->buffer_cont_size)
+			if (/*self->buffer_read_size == 0 || > condition not necessary */
+					self->buffer_pos == self->buffer_read_size)
 			{
 				int res;
 				self->buffer_pos = 0;
-				self->buffer_cont_size = 0;
+				self->buffer_read_size = 0;
 				/* Note: this condition is important, you cannot rely on
 				 * having a call to RARProcessFileChunk return no data and
 				 * break on the condition self->buffer_cont_size == 0 because
@@ -128,25 +133,25 @@ static size_t php_rar_ops_read(php_stream *stream, char *buf, size_t count TSRML
 				if (self->no_more_data)
 					break;
 				res = RARProcessFileChunk(self->rar_handle, self->buffer,
-					self->buffer_size, &self->buffer_cont_size,
+						self->buffer_size, &self->buffer_read_size,
 					&self->no_more_data);
 				if (_rar_handle_error(res TSRMLS_CC) == FAILURE) {
 					break; /* finish in case of failure */
 				}
-				assert(self->buffer_cont_size <= self->buffer_size);
+				assert(self->buffer_read_size <= self->buffer_size);
 				/* we did not read anything. no need to continue */
-				if (self->buffer_cont_size == 0)
+				if (self->buffer_read_size == 0)
 					break;
 			}
 			/* if we get here we have data to be read in the buffer */
 			this_read_size = MIN(left,
-				self->buffer_cont_size - self->buffer_pos);
+					self->buffer_read_size - self->buffer_pos);
 			assert(this_read_size > 0);
 			memcpy(&buf[count-left], &self->buffer[self->buffer_pos],
 				this_read_size);
 			left				-= this_read_size;
 			n					+= this_read_size;
-			self->buffer_pos		+= this_read_size;
+			self->buffer_pos	+= this_read_size;
 			assert(left >= 0);
 		}
 
@@ -155,9 +160,16 @@ static size_t php_rar_ops_read(php_stream *stream, char *buf, size_t count TSRML
 
 	/* no more data upstream (for sure), buffer already read and
 	 * caller asked for more data than we're giving */
-	if (self->no_more_data && self->buffer_pos == self->buffer_cont_size &&
-			((size_t) n) < count)
+	if (self->no_more_data && self->buffer_pos == self->buffer_read_size &&
+			n < count && stream->eof != 1) {
 		stream->eof = 1;
+		if (self->cursor > self->file_size) {
+			php_error_docref(NULL TSRMLS_CC, E_WARNING,
+					"The file size is supposed to be %lu bytes, but "
+					"we read more: %lu bytes (corruption/wrong pwd)",
+					self->file_size, self->cursor);
+		}
+	}
 
 	/* we should only give no data if we have no more */
 	if (!self->no_more_data && n == 0) {
@@ -167,17 +179,29 @@ static size_t php_rar_ops_read(php_stream *stream, char *buf, size_t count TSRML
 		stream->eof = 1;
 	}
 
+#if PHP_VERSION_ID < 50400
 	return n;
+#else
+	return (ssize_t) n;
+#endif
 }
 /* }}} */
 
 /* {{{ php_rar_ops_write */
+#if PHP_VERSION_ID < 70400
 static size_t php_rar_ops_write(php_stream *stream, const char *buf, size_t count TSRMLS_DC)
+#else
+static ssize_t php_rar_ops_write(php_stream *stream, const char *buf, size_t count TSRMLS_DC)
+#endif
 {
 	php_error_docref(NULL TSRMLS_CC, E_WARNING,
 		"Write operation not supported for RAR streams.");
 	if (!stream) {
+#if PHP_VERSION_ID < 70400
 		return 0;
+#else
+		return -1;
+#endif
 	}
 
 	return count;
@@ -293,7 +317,7 @@ static int _rar_stat_from_header(struct RARHeaderDataEx *header,
 	 * SUBHEAD_TYPE_UOWNER), but it is not exposed in unRAR */
 	ssb->sb.st_uid = 0;
 	ssb->sb.st_gid = 0;
-#ifdef HAVE_ST_RDEV
+#if defined(HAVE_ST_RDEV) || defined(HAVE_STRUCT_STAT_ST_RDEV)
 	ssb->sb.st_rdev = 0;
 #endif
 	/* never mind signedness, we'll never get sizes big enough for that to
@@ -324,10 +348,10 @@ static int _rar_stat_from_header(struct RARHeaderDataEx *header,
 			&ssb->sb.st_mtime);
 	}
 
-#ifdef HAVE_ST_BLKSIZE
+#if defined(HAVE_ST_BLKSIZE) || defined(HAVE_STRUCT_STAT_ST_BLKSIZE)
 	ssb->sb.st_blksize = 0;
 #endif
-#ifdef HAVE_ST_BLOCKS
+#if defined(HAVE_ST_BLOCKS) || defined (HAVE_STRUCT_STAT_ST_BLOCKS)
 	ssb->sb.st_blocks = 0;
 #endif
 	/* php_stat in filestat.c doesn't check this one, so don't touch it */
@@ -363,19 +387,32 @@ static php_stream_ops php_stream_rario_ops = {
 /* {{{ RAR directory streams */
 
 /* {{{ php_rar_dir_ops_read */
+#if PHP_VERSION_ID < 70400
 static size_t php_rar_dir_ops_read(php_stream *stream, char *buf, size_t count TSRMLS_DC)
+#else
+static ssize_t php_rar_dir_ops_read(php_stream *stream, char *buf, size_t count TSRMLS_DC)
+#endif
 {
 	php_stream_dirent entry;
-	int offset;
+	size_t offset;
 	STREAM_DIR_DATA_FROM_STREAM
 
-	if (count != sizeof(entry))
+	if (count != sizeof(entry)) {
+#if PHP_VERSION_ID < 70400
 		return 0;
+#else
+		return -1;
+#endif
+	}
 
 	_rar_entry_search_advance(self->state, self->directory, self->dir_size, 1);
 	if (!self->state->found) {
 		stream->eof = 1;
+#if PHP_VERSION_ID < 70400
 		return 0;
+#else
+		return -1;
+#endif
 	}
 
 	if (self->dir_size == 1) /* root */
@@ -431,7 +468,11 @@ static int php_rar_dir_ops_close(php_stream *stream, int close_handle TSRMLS_DC)
 /* }}} */
 
 /* {{{ php_rar_dir_ops_rewind */
+#if PHP_VERSION_ID < 70400
 static int php_rar_dir_ops_rewind(php_stream *stream, off_t offset, int whence, off_t *newoffset TSRMLS_DC)
+#else
+static int php_rar_dir_ops_rewind(php_stream *stream, zend_off_t offset, int whence, zend_off_t *newoffset)
+#endif
 {
 	STREAM_DIR_DATA_FROM_STREAM
 
@@ -508,16 +549,18 @@ php_stream *php_stream_rar_open(char *arc_name,
 			TSRMLS_CC, position, arc_name);
 	else {
 		/* no need to allocate a buffer bigger than the file uncomp size */
-		size_t buffer_size = (size_t)
-			MIN((uint64) RAR_CHUNK_BUFFER_SIZE,
-			INT32TO64(self->header_data.UnpSizeHigh,
-			self->header_data.UnpSize));
+		size_t file_size = INT32TO64(self->header_data.UnpSizeHigh,
+				self->header_data.UnpSize);
+		size_t buffer_size =  MIN(
+				MAX(RAR_CHUNK_BUFFER_SIZE, self->header_data.WinSize),
+				file_size);
 		int process_result = RARProcessFileChunkInit(self->rar_handle);
 
 		if (_rar_handle_error(process_result TSRMLS_CC) == FAILURE) {
 			goto cleanup;
 		}
 
+		self->file_size = file_size;
 		self->buffer = emalloc(buffer_size);
 		self->buffer_size = buffer_size;
 		stream = php_stream_alloc(&php_stream_rario_ops, self, NULL, "rb");
@@ -909,21 +952,23 @@ static php_stream *php_stream_rar_opener(php_stream_wrapper *wrapper,
 
 	{
 		/* no need to allocate a buffer bigger than the file uncomp size */
-		size_t buffer_size = (size_t)
-			MIN((uint64) RAR_CHUNK_BUFFER_SIZE,
-			INT32TO64(self->header_data.UnpSizeHigh,
-			self->header_data.UnpSize));
+		size_t file_size = INT32TO64(self->header_data.UnpSizeHigh,
+				self->header_data.UnpSize);
+		size_t buffer_size =  MIN(
+				MAX(RAR_CHUNK_BUFFER_SIZE, self->header_data.WinSize),
+				file_size);
 		rar_result = RARProcessFileChunkInit(self->rar_handle);
 
 		if ((rar_error = _rar_error_to_string(rar_result)) != NULL) {
 			char *mb_entry = _rar_wide_to_utf_with_alloc(fragment, -1);
 			php_stream_wrapper_log_error(wrapper, options TSRMLS_CC,
-				"Error opening file %s inside RAR archive %s: %s",
-				mb_entry, tmp_open_path, rar_error);
+					"Error opening file %s inside RAR archive %s: %s",
+					mb_entry, tmp_open_path, rar_error);
 			efree(mb_entry);
 			goto cleanup;
 		}
 
+		self->file_size = file_size;
 		self->buffer = emalloc(buffer_size);
 		self->buffer_size = buffer_size;
 		stream = php_stream_alloc(&php_stream_rario_ops, self, NULL, mode);
