@@ -41,23 +41,20 @@ extern "C" {
 zend_class_entry *rar_class_entry_ptr;
 /* }}} */
 
-/* {{{ Globals with internal linkage */
-static zend_object_handlers rarentry_object_handlers;
-/* }}} */
-
 /* {{{ Function prototypes for functions with internal linkage */
 static int _rar_decl_priv_prop_null(zend_class_entry *ce, const char *name,
 									 int name_length, char *doc_comment,
 									 int doc_comment_len TSRMLS_DC);
 static zval *_rar_entry_get_property(zval *entry_obj, char *name, int namelen TSRMLS_DC);
-static void _rar_dos_date_to_text(int dos_time, char *date_string);
-static zend_object_value rarentry_ce_create_object(zend_class_entry *class_type TSRMLS_DC);
+static void _rar_dos_date_to_text(unsigned dos_time, char *date_string);
 /* }}} */
 
 /* {{{ Functions with external linkage */
 /* should be passed the last entry that corresponds to a given file
  * only that one has the correct CRC. Still, it may have a wrong packedSize */
-void _rar_entry_to_zval(zval *parent, /* zval to RarArchive object, will have its refcount increased */
+/* parent is zval to RarArchive object. The object (not the zval, in PHP 5.x)
+ * will have its refcount increased */
+void _rar_entry_to_zval(zval *parent,
 						struct RARHeaderDataEx *entry,
 						unsigned long packed_size,
 						size_t position,
@@ -66,12 +63,21 @@ void _rar_entry_to_zval(zval *parent, /* zval to RarArchive object, will have it
 	char tmp_s [MAX_LENGTH_OF_LONG + 1];
 	char time[50];
 	char *filename;
-	int  filename_size, filename_len;
+	int  filename_size,
+		 filename_len;
 	long unp_size; /* zval stores PHP ints as long, so use that here */
+	zval *parent_copy = parent;
+#if PHP_MAJOR_VERSION < 7
+	/* allocate zval on the heap */
+	zval_addref_p(parent_copy);
+	SEPARATE_ZVAL(&parent_copy);
+	/* set refcount to 0; zend_update_property will increase it */
+	Z_DELREF_P(parent_copy);
+#endif
 
 	object_init_ex(object, rar_class_entry_ptr);
 	zend_update_property(rar_class_entry_ptr, object, "rarfile",
-		sizeof("rararch") - 1, parent TSRMLS_CC);
+		sizeof("rararch") - 1, parent_copy TSRMLS_CC);
 
 #if ULONG_MAX > 0xffffffffUL
 	unp_size = ((long) entry->UnpSize) + (((long) entry->UnpSizeHigh) << 32);
@@ -84,7 +90,7 @@ void _rar_entry_to_zval(zval *parent, /* zval to RarArchive object, will have it
 		unp_size = (long) entry->UnpSize;
 #endif
 
-	filename_size = sizeof(entry->FileNameW) * sizeof(wchar_t);
+	filename_size = sizeof(entry->FileNameW) * 4;
 	filename = (char*) emalloc(filename_size);
 
 	if (packed_size > (unsigned long) LONG_MAX)
@@ -124,6 +130,28 @@ void _rar_entry_to_zval(zval *parent, /* zval to RarArchive object, will have it
 	zend_update_property_long(rar_class_entry_ptr, object, "flags",
 		sizeof("flags") - 1, entry->Flags TSRMLS_CC);
 
+	zend_update_property_long(rar_class_entry_ptr, object, "redir_type",
+		sizeof("redir_type") - 1, entry->RedirType TSRMLS_CC);
+
+	if (entry->RedirName) {
+		char *redir_target = NULL;
+		size_t redir_target_size;
+
+		zend_update_property_bool(rar_class_entry_ptr, object,
+			"redir_to_directory", sizeof("redir_to_directory") - 1,
+			!!entry->DirTarget TSRMLS_CC);
+
+		redir_target_size = entry->RedirNameSize * 4;
+		redir_target =  emalloc(redir_target_size);
+		assert(redir_target_size > 0);
+		_rar_wide_to_utf(entry->RedirName, redir_target, redir_target_size);
+
+		zend_update_property_string(rar_class_entry_ptr, object, "redir_target",
+			sizeof("redir_target") - 1, redir_target TSRMLS_CC);
+
+		efree(redir_target);
+	}
+
 	efree(filename);
 }
 /* }}} */
@@ -152,67 +180,76 @@ static int _rar_decl_priv_prop_null(zend_class_entry *ce, const char *name,
 									 int name_length, char *doc_comment,
 									 int doc_comment_len TSRMLS_DC) /* {{{ */
 {
+#if PHP_MAJOR_VERSION < 7
 	zval *property;
 	ALLOC_PERMANENT_ZVAL(property);
 	INIT_ZVAL(*property);
 	return zend_declare_property_ex(ce, name, name_length, property,
 		ZEND_ACC_PRIVATE, doc_comment, doc_comment_len TSRMLS_CC);
+#else
+	zval property;
+	zend_string *name_str,
+				*doc_str;
+	int ret;
+
+	ZVAL_NULL(&property);
+	name_str = zend_string_init(name, (size_t) name_length, 1);
+	doc_str = zend_string_init(doc_comment, (size_t) doc_comment_len, 1);
+	ret = zend_declare_property_ex(ce, name_str, &property, ZEND_ACC_PRIVATE,
+								   doc_str);
+	zend_string_release(name_str);
+	zend_string_release(doc_str);
+	return ret;
+#endif
 }
 /* }}} */
 
 static zval *_rar_entry_get_property(zval *entry_obj, char *name, int namelen TSRMLS_DC) /* {{{ */
 {
 	zval *tmp;
+#if PHP_MAJOR_VERSION >= 7
+	zval zv;
+#endif
+#if PHP_VERSION_ID < 70100
 	zend_class_entry *orig_scope = EG(scope);
 
 	EG(scope) = rar_class_entry_ptr;
+#endif
 
+#if PHP_MAJOR_VERSION < 7
 	tmp = zend_read_property(Z_OBJCE_P(entry_obj), entry_obj, name, namelen, 1 TSRMLS_CC);
+#else
+	tmp = zend_read_property(Z_OBJCE_P(entry_obj), entry_obj, name, namelen, 1, &zv);
+#endif
 	if (tmp == NULL) {
 		php_error_docref(NULL TSRMLS_CC, E_WARNING,
 			"Bug: unable to find property '%s'. Please report.", name);
 	}
 
+#if PHP_VERSION_ID < 70100
 	EG(scope) = orig_scope;
+#endif
 
 	return tmp;
 }
 /* }}} */
 
-static void _rar_dos_date_to_text(int dos_time, char *date_string) /* {{{ */
+static void _rar_dos_date_to_text(unsigned dos_time, char *date_string) /* {{{ */
 {
-	int second, minute, hour, day, month, year;
-	/* following lines were taken from timefn.cpp */
-	second = (dos_time & 0x1f)*2;
-	minute = (dos_time>>5) & 0x3f;
-	hour   = (dos_time>>11) & 0x1f;
-	day    = (dos_time>>16) & 0x1f;
-	month  = (dos_time>>21) & 0x0f;
-	year   = (dos_time>>25)+1980;
-	sprintf(date_string, "%u-%02u-%02u %02u:%02u:%02u", year, month, day, hour, minute, second);
-}
-/* }}} */
+	time_t time = 0;
+	struct tm tm = {0};
+	int res;
 
-static zend_object_value rarentry_ce_create_object(zend_class_entry *class_type TSRMLS_DC) /* {{{ */
-{
-	zend_object_value	zov;
-	zend_object			*zobj;
+	res = rar_dos_time_convert(dos_time, &time) != FAILURE &&
+		php_gmtime_r(&time, &tm) != NULL;
 
-	zobj = emalloc(sizeof *zobj);
-	zend_object_std_init(zobj, class_type TSRMLS_CC);
+	if (!res) {
+		sprintf(date_string, "%s", "time conversion failure");
+	}
 
-#if PHP_VERSION_ID < 50399
-	zend_hash_copy(zobj->properties, &(class_type->default_properties),
-		(copy_ctor_func_t) zval_add_ref, NULL, sizeof(zval*));
-#else
-	object_properties_init(zobj, class_type);
-#endif
-	zov.handle = zend_objects_store_put(zobj,
-		(zend_objects_store_dtor_t) zend_objects_destroy_object,
-		(zend_objects_free_object_storage_t) zend_objects_free_object_storage,
-		NULL TSRMLS_CC);
-	zov.handlers = &rarentry_object_handlers;
-	return zov;
+	sprintf(date_string, "%u-%02u-%02u %02u:%02u:%02u",
+		tm.tm_year + 1900, tm.tm_mon + 1, tm.tm_mday, tm.tm_hour, tm.tm_min,
+		tm.tm_sec);
 }
 /* }}} */
 /* }}} */
@@ -226,7 +263,7 @@ PHP_METHOD(rarentry, extract)
 	char					*dir,
 							*filepath = NULL,
 							*password = NULL;
-	int						dir_len,
+	zpp_s_size_t			dir_len,
 							filepath_len = 0,
 							password_len = 0;
 	char					*considered_path;
@@ -360,7 +397,7 @@ PHP_METHOD(rarentry, getName)
 
 	RAR_GET_PROPERTY(tmp, "name");
 
-	RETURN_STRINGL(Z_STRVAL_P(tmp), Z_STRLEN_P(tmp), 1);
+	RAR_RETURN_STRINGL(Z_STRVAL_P(tmp), Z_STRLEN_P(tmp), 1);
 }
 /* }}} */
 
@@ -410,7 +447,9 @@ PHP_METHOD(rarentry, getHostOs)
 /* }}} */
 
 /* {{{ proto string RarEntry::getFileTime()
-   Return modification time of the entry */
+   Return modification time of the entry.
+   Due to the way the unrar library returns this time, this is in the
+   system's timezone. */
 PHP_METHOD(rarentry, getFileTime)
 {
 	zval *tmp;
@@ -420,7 +459,7 @@ PHP_METHOD(rarentry, getFileTime)
 
 	RAR_GET_PROPERTY(tmp, "file_time");
 
-	RETURN_STRINGL(Z_STRVAL_P(tmp), Z_STRLEN_P(tmp), 1);
+	RAR_RETURN_STRINGL(Z_STRVAL_P(tmp), Z_STRLEN_P(tmp), 1);
 }
 /* }}} */
 
@@ -435,7 +474,7 @@ PHP_METHOD(rarentry, getCrc)
 
 	RAR_GET_PROPERTY(tmp, "crc");
 
-	RETURN_STRINGL(Z_STRVAL_P(tmp), Z_STRLEN_P(tmp), 1);
+	RAR_RETURN_STRINGL(Z_STRVAL_P(tmp), Z_STRLEN_P(tmp), 1);
 }
 /* }}} */
 
@@ -494,7 +533,7 @@ PHP_METHOD(rarentry, getStream)
 	zval				*entry_obj = getThis();
 	php_stream			*stream = NULL;
 	char				*password = NULL;
-	int					password_len; /* ignored */
+	zpp_s_size_t		password_len; /* ignored */
 	rar_cb_user_data	cb_udata = {NULL};
 
 
@@ -539,7 +578,7 @@ PHP_METHOD(rarentry, isDirectory)
 
 	RAR_GET_PROPERTY(tmp, "flags");
 	flags = Z_LVAL_P(tmp);
-	is_dir = ((flags & LHD_WINDOWMASK) == LHD_DIRECTORY);
+	is_dir = (flags & RHDF_DIRECTORY) != 0;
 
 	RETURN_BOOL(is_dir);
 }
@@ -558,12 +597,66 @@ PHP_METHOD(rarentry, isEncrypted)
 
 	RAR_GET_PROPERTY(tmp, "flags");
 	flags = Z_LVAL_P(tmp);
-	is_encrypted = (flags & 0x04);
+	is_encrypted = (flags & RHDF_ENCRYPTED) != 0;
 
 	RETURN_BOOL(is_encrypted);
 }
 /* }}} */
 
+/* {{{ proto int RarEntry::getRedirType()
+   Returns the redirection type, or NULL if there's none */
+PHP_METHOD(rarentry, getRedirType)
+{
+	zval *tmp;
+	zval *entry_obj = getThis();
+
+	RAR_RETNULL_ON_ARGS();
+
+	RAR_GET_PROPERTY(tmp, "redir_type");
+	if (Z_TYPE_P(tmp) != IS_LONG) {
+		php_error_docref(NULL TSRMLS_CC, E_WARNING, "bad redir type stored");
+		RETURN_FALSE;
+	}
+
+	if (Z_LVAL_P(tmp) == FSREDIR_NONE) {
+		RETURN_NULL();
+	}
+
+	RETURN_LONG(Z_LVAL_P(tmp));
+}
+/* }}} */
+
+/* {{{ proto bool RarEntry::isRedirectToDirectory()
+   Returns true if there is redirection and the target is a directory,
+   null if there is no redirection, false otherwise */
+PHP_METHOD(rarentry, isRedirectToDirectory)
+{
+	zval *tmp;
+	zval *entry_obj = getThis();
+
+	RAR_RETNULL_ON_ARGS();
+
+	RAR_GET_PROPERTY(tmp, "redir_to_directory");
+
+	RETURN_ZVAL(tmp, 1, 0);
+}
+/* }}} */
+
+/* {{{ proto bool RarEntry::getRedirTarget()
+   Returns the redirection target, encoded as UTF-8, or NULL */
+PHP_METHOD(rarentry, getRedirTarget)
+{
+	zval *tmp;
+	zval *entry_obj = getThis();
+
+	RAR_RETNULL_ON_ARGS();
+
+	RAR_GET_PROPERTY(tmp, "redir_target");
+
+	RETURN_ZVAL(tmp, 1, 0);
+}
+/* }}} */
+
 /* {{{ proto string RarEntry::__toString()
    Return string representation for entry */
 PHP_METHOD(rarentry, __toString)
@@ -577,14 +670,14 @@ PHP_METHOD(rarentry, __toString)
 	char		*name,
 				*crc;
 	char		*restring;
-	int			restring_len;
+	int			restring_size;
 	const char	format[] = "RarEntry for %s \"%s\" (%s)";
 
 	RAR_RETNULL_ON_ARGS();
 
 	RAR_GET_PROPERTY(flags_zval, "flags");
 	flags = Z_LVAL_P(flags_zval);
-	is_dir = ((flags & 0xE0) == 0xE0);
+	is_dir = flags & RHDF_DIRECTORY;
 
 	RAR_GET_PROPERTY(name_zval, "name");
 	name = Z_STRVAL_P(name_zval);
@@ -593,14 +686,14 @@ PHP_METHOD(rarentry, __toString)
 	crc = Z_STRVAL_P(crc_zval);
 
 	/* 2 is size of %s, 8 is size of crc */
-	restring_len = (sizeof(format)-1) - 2 * 3 + (sizeof("directory")-1) +
+	restring_size = (sizeof(format)-1) - 2 * 3 + (sizeof("directory")-1) +
 		strlen(name) + 8 + 1;
-	restring = emalloc(restring_len);
-	snprintf(restring, restring_len, format, is_dir?"directory":"file",
+	restring = emalloc(restring_size);
+	snprintf(restring, restring_size, format, is_dir?"directory":"file",
 		name, crc);
-	restring[restring_len - 1] = '\0'; /* just to be safe */
+	restring[restring_size - 1] = '\0'; /* just to be safe */
 
-	RETURN_STRING(restring, 0);
+	RAR_RETURN_STRINGL(restring, strlen(restring), 0);
 }
 /* }}} */
 /* }}} */
@@ -636,6 +729,9 @@ static zend_function_entry php_rar_class_functions[] = {
 	PHP_ME(rarentry,		getStream,			arginfo_rarentry_getstream,	ZEND_ACC_PUBLIC)
 	PHP_ME(rarentry,		isDirectory,		arginfo_rar_void,	ZEND_ACC_PUBLIC)
 	PHP_ME(rarentry,		isEncrypted,		arginfo_rar_void,	ZEND_ACC_PUBLIC)
+	PHP_ME(rarentry,		getRedirType,		arginfo_rar_void,	ZEND_ACC_PUBLIC)
+	PHP_ME(rarentry,		isRedirectToDirectory,	arginfo_rar_void,	ZEND_ACC_PUBLIC)
+	PHP_ME(rarentry,		getRedirTarget,	arginfo_rar_void,	ZEND_ACC_PUBLIC)
 	PHP_ME(rarentry,		__toString,			arginfo_rar_void,	ZEND_ACC_PUBLIC)
 	PHP_ME_MAPPING(__construct,	rar_bogus_ctor,	arginfo_rar_void,	ZEND_ACC_PRIVATE | ZEND_ACC_CTOR)
 	{NULL, NULL, NULL}
@@ -645,15 +741,10 @@ void minit_rarentry(TSRMLS_D)
 {
 	zend_class_entry ce;
 
-	memcpy(&rarentry_object_handlers, zend_get_std_object_handlers(),
-		sizeof rarentry_object_handlers);
-
 	INIT_CLASS_ENTRY(ce, "RarEntry", php_rar_class_functions);
 	rar_class_entry_ptr = zend_register_internal_class(&ce TSRMLS_CC);
 	rar_class_entry_ptr->ce_flags |= ZEND_ACC_FINAL_CLASS;
 	rar_class_entry_ptr->clone = NULL;
-	/* Custom creation currently not really needed, but you never know... */
-	rar_class_entry_ptr->create_object = &rarentry_ce_create_object;
 
 	REG_RAR_PROPERTY("rarfile", "Associated RAR archive");
 	REG_RAR_PROPERTY("position", "Position inside the RAR archive");
@@ -667,6 +758,9 @@ void minit_rarentry(TSRMLS_D)
 	REG_RAR_PROPERTY("version", "RAR version needed to extract entry");
 	REG_RAR_PROPERTY("method", "Identifier for packing method");
 	REG_RAR_PROPERTY("flags", "Entry header flags");
+	REG_RAR_PROPERTY("redir_type", "The type of redirection or NULL");
+	REG_RAR_PROPERTY("redir_to_directory", "Whether the redirection target is a directory");
+	REG_RAR_PROPERTY("redir_target", "Target of the redirectory");
 
 	REG_RAR_CLASS_CONST_LONG("HOST_MSDOS",	HOST_MSDOS);
 	REG_RAR_CLASS_CONST_LONG("HOST_OS2",	HOST_OS2);
@@ -675,6 +769,12 @@ void minit_rarentry(TSRMLS_D)
 	REG_RAR_CLASS_CONST_LONG("HOST_MACOS",	HOST_MACOS);
 	REG_RAR_CLASS_CONST_LONG("HOST_BEOS",	HOST_BEOS);
 
+	REG_RAR_CLASS_CONST_LONG("FSREDIR_UNIXSYMLINK",	FSREDIR_UNIXSYMLINK);
+	REG_RAR_CLASS_CONST_LONG("FSREDIR_WINSYMLINK",	FSREDIR_WINSYMLINK);
+	REG_RAR_CLASS_CONST_LONG("FSREDIR_JUNCTION",	FSREDIR_JUNCTION);
+	REG_RAR_CLASS_CONST_LONG("FSREDIR_HARDLINK",	FSREDIR_HARDLINK);
+	REG_RAR_CLASS_CONST_LONG("FSREDIR_FILECOPY",	FSREDIR_FILECOPY);
+
 	/* see WinNT.h */
 	REG_RAR_CLASS_CONST_LONG("ATTRIBUTE_WIN_READONLY",				0x00001L);
 	REG_RAR_CLASS_CONST_LONG("ATTRIBUTE_WIN_HIDDEN",				0x00002L);
