@@ -35,11 +35,14 @@ HANDLE PASCAL RAROpenArchiveEx(struct RAROpenArchiveDataEx *r)
   DataSet *Data=NULL;
   try
   {
+    ErrHandler.Clean();
+
     r->OpenResult=0;
     Data=new DataSet;
     Data->Cmd.DllError=0;
     Data->OpenMode=r->OpenMode;
     Data->Cmd.FileArgs.AddString(L"*");
+    Data->Cmd.KeepBroken=(r->OpFlags&ROADOF_KEEPBROKEN)!=0;
 
     char AnsiArcName[NM];
     *AnsiArcName=0;
@@ -48,7 +51,7 @@ HANDLE PASCAL RAROpenArchiveEx(struct RAROpenArchiveDataEx *r)
       strncpyz(AnsiArcName,r->ArcName,ASIZE(AnsiArcName));
 #ifdef _WIN_ALL
       if (!AreFileApisANSI())
-    {
+      {
         OemToCharBuffA(r->ArcName,AnsiArcName,ASIZE(AnsiArcName));
         AnsiArcName[ASIZE(AnsiArcName)-1]=0;
       }
@@ -90,39 +93,53 @@ HANDLE PASCAL RAROpenArchiveEx(struct RAROpenArchiveDataEx *r)
       return NULL;
     }
     r->Flags=0;
-
+    
     if (Data->Arc.Volume)
-      r->Flags|=0x01;
+      r->Flags|=ROADF_VOLUME;
+    if (Data->Arc.MainComment)
+      r->Flags|=ROADF_COMMENT;
     if (Data->Arc.Locked)
-      r->Flags|=0x04;
+      r->Flags|=ROADF_LOCK;
     if (Data->Arc.Solid)
-      r->Flags|=0x08;
+      r->Flags|=ROADF_SOLID;
     if (Data->Arc.NewNumbering)
-      r->Flags|=0x10;
+      r->Flags|=ROADF_NEWNUMBERING;
     if (Data->Arc.Signed)
-      r->Flags|=0x20;
+      r->Flags|=ROADF_SIGNED;
     if (Data->Arc.Protected)
-      r->Flags|=0x40;
+      r->Flags|=ROADF_RECOVERY;
     if (Data->Arc.Encrypted)
-      r->Flags|=0x80;
+      r->Flags|=ROADF_ENCHEADERS;
     if (Data->Arc.FirstVolume)
-      r->Flags|=0x100;
+      r->Flags|=ROADF_FIRSTVOLUME;
 
     Array<wchar> CmtDataW;
     if (r->CmtBufSize!=0 && Data->Arc.GetComment(&CmtDataW))
     {
+      if (r->CmtBufW!=NULL)
+      {
+        CmtDataW.Push(0);
+        size_t Size=wcslen(&CmtDataW[0])+1;
+
+        r->CmtState=Size>r->CmtBufSize ? ERAR_SMALL_BUF:1;
+        r->CmtSize=(uint)Min(Size,r->CmtBufSize);
+        memcpy(r->CmtBufW,&CmtDataW[0],(r->CmtSize-1)*sizeof(*r->CmtBufW));
+        r->CmtBufW[r->CmtSize-1]=0;
+      }
+      else
+        if (r->CmtBuf!=NULL)
+        {
       Array<char> CmtData(CmtDataW.Size()*4+1);
       memset(&CmtData[0],0,CmtData.Size());
       WideToChar(&CmtDataW[0],&CmtData[0],CmtData.Size()-1);
       size_t Size=strlen(&CmtData[0])+1;
 
-      r->Flags|=2;
       r->CmtState=Size>r->CmtBufSize ? ERAR_SMALL_BUF:1;
       r->CmtSize=(uint)Min(Size,r->CmtBufSize);
       memcpy(r->CmtBuf,&CmtData[0],r->CmtSize-1);
-      if (Size<=r->CmtBufSize)
         r->CmtBuf[r->CmtSize-1]=0;
     }
+    }
     else
       r->CmtState=r->CmtSize=0;
     Data->Extract.ExtractArchiveInit(Data->Arc);
@@ -151,10 +168,17 @@ HANDLE PASCAL RAROpenArchiveEx(struct RAROpenArchiveDataEx *r)
 int PASCAL RARCloseArchive(HANDLE hArcData)
 {
   DataSet *Data=(DataSet *)hArcData;
+  try
+  {
   bool Success=Data==NULL ? false:Data->Arc.Close();
   delete Data;
   return Success ? ERAR_SUCCESS : ERAR_ECLOSE;
 }
+  catch (RAR_EXIT ErrCode)
+  {
+    return Data->Cmd.DllError!=0 ? Data->Cmd.DllError : RarErrorToDll(ErrCode);
+  }
+}
 
 
 int PASCAL RARReadHeader(HANDLE hArcData,struct RARHeaderData *D)
@@ -243,9 +267,6 @@ int PASCAL RARReadHeaderEx(HANDLE hArcData,struct RARHeaderDataEx *D)
     D->UnpSize=uint(hd->UnpSize & 0xffffffff);
     D->UnpSizeHigh=uint(hd->UnpSize>>32);
     D->HostOS=hd->HSType==HSYS_WINDOWS ? HOST_WIN32:HOST_UNIX;
-    if (Data->Arc.Format==RARFMT50)
-      D->UnpVer=Data->Arc.FileHead.UnpVer==0 ? 50 : 200; // If it is not 0, just set it to something big.
-    else
       D->UnpVer=Data->Arc.FileHead.UnpVer;
     D->FileCRC=hd->FileHash.CRC32;
     D->FileTime=hd->mtime.GetDos();
@@ -291,6 +312,9 @@ int PASCAL RARReadHeaderEx(HANDLE hArcData,struct RARHeaderDataEx *D)
         D->RedirNameSize>0 && D->RedirNameSize<100000)
       wcsncpyz(D->RedirName,hd->RedirName,D->RedirNameSize);
     D->DirTarget=hd->DirTarget;
+
+    /* added by me */
+    D->WinSize = hd->WinSize;
   }
   catch (RAR_EXIT ErrCode)
   {
@@ -379,7 +403,7 @@ int PASCAL ProcessFile(HANDLE hArcData, int Operation, char *DestPath,
       if (DestNameW!=NULL)
         wcsncpyz(Data->Cmd.DllDestName,DestNameW,ASIZE(Data->Cmd.DllDestName));
 
-      wcscpy(Data->Cmd.Command,Operation==RAR_EXTRACT ? L"X":L"T");
+      wcsncpyz(Data->Cmd.Command,Operation==RAR_EXTRACT ? L"X":L"T",ASIZE(Data->Cmd.Command));
       Data->Cmd.Test=Operation!=RAR_EXTRACT;
       if (Operation == RAR_EXTRACT_CHUNK)
       {
@@ -393,7 +417,7 @@ int PASCAL ProcessFile(HANDLE hArcData, int Operation, char *DestPath,
 
       bool Repeat=false;
       if (Operation != RAR_EXTRACT_CHUNK)
-        Data->Extract.ExtractCurrentFile(Data->Arc,Data->HeaderSize,Repeat);
+          Data->Extract.ExtractCurrentFile(Data->Arc,Data->HeaderSize,Repeat);
       else
       {
         if (InitDataIO) //chunk, init
@@ -432,7 +456,7 @@ int PASCAL ProcessFile(HANDLE hArcData, int Operation, char *DestPath,
       }
       Data->Arc.Seek(Data->Arc.CurBlockPos,SEEK_SET);
     }
-    }
+  }
   }
   catch (std::bad_alloc&)
   {
@@ -448,15 +472,15 @@ int PASCAL ProcessFile(HANDLE hArcData, int Operation, char *DestPath,
 
 int PASCAL RARProcessFile(HANDLE hArcData,int Operation,char *DestPath,char *DestName)
 {
-  return(ProcessFile(hArcData,Operation,DestPath,DestName,NULL,NULL,NULL,0,
-    NULL,false,NULL));
+  return ProcessFile(hArcData,Operation,DestPath,DestName,NULL,NULL,NULL,0,
+    NULL,false,NULL);
 }
 
 
 int PASCAL RARProcessFileW(HANDLE hArcData,int Operation,wchar *DestPath,wchar *DestName)
 {
-  return(ProcessFile(hArcData,Operation,NULL,NULL,DestPath,DestName,NULL,0,
-    NULL,false,NULL));
+  return ProcessFile(hArcData,Operation,NULL,NULL,DestPath,DestName,NULL,0,
+    NULL,false,NULL);
 }
 
 int PASCAL RARProcessFileChunkInit(HANDLE hArcData)
@@ -505,16 +529,16 @@ void PASCAL RARSetProcessDataProc(HANDLE hArcData,PROCESSDATAPROC ProcessDataPro
 }
 
 
-#ifndef RAR_NOCRYPT
 void PASCAL RARSetPassword(HANDLE hArcData,char *Password)
 {
+#ifndef RAR_NOCRYPT
   DataSet *Data=(DataSet *)hArcData;
   wchar PasswordW[MAXPASSWORD];
   GetWideName(Password,NULL,PasswordW,ASIZE(PasswordW));
   Data->Cmd.Password.Set(PasswordW);
   cleandata(PasswordW,sizeof(PasswordW));
-}
 #endif
+}
 
 
 int PASCAL RARGetDllVersion()
