@@ -1,7 +1,7 @@
 #ifndef _UNRAR_DLL_
 #define _UNRAR_DLL_
 
-#pragma pack(1)
+#pragma pack(push, 1)
 
 #define ERAR_SUCCESS             0
 #define ERAR_END_ARCHIVE        10
@@ -39,9 +39,11 @@
 #define RAR_HASH_CRC32        1
 #define RAR_HASH_BLAKE2       2
 
-//Must be the same as MAXWINSIZE
 //not in original
-#define RAR_CHUNK_BUFFER_SIZE   0x400000
+// should be a multiple of 16 becasue this buffer is passed
+// directly to UnpRead (bypassing UnstoreFile) and 16 is the size
+// of crypt blocks
+#define RAR_CHUNK_BUFFER_SIZE   0x100000UL
 
 #ifdef _UNIX
 #define CALLBACK
@@ -131,6 +133,8 @@ struct RARHeaderDataEx
   /* removed by me: we don't need to retain binary compatibility in case new
    * fields are added, so we avoid wasting space here */
   /* unsigned int Reserved[988]; */
+  /* added by me */
+  size_t WinSize; /* window size */
 };
 
 
@@ -157,6 +161,8 @@ typedef int (CALLBACK *UNRARCALLBACK)(UINT msg,LPARAM UserData,LPARAM P1,LPARAM
 #define ROADF_ENCHEADERS   0x0080
 #define ROADF_FIRSTVOLUME  0x0100
 
+#define ROADOF_KEEPBROKEN  0x0001
+
 struct RAROpenArchiveDataEx
 {
   char         *ArcName;
@@ -170,8 +176,10 @@ struct RAROpenArchiveDataEx
   unsigned int  Flags;
   UNRARCALLBACK Callback;
   LPARAM        UserData;
+  unsigned int  OpFlags;
+  wchar_t      *CmtBufW;
   /* removed by me */
-  /* unsigned int  Reserved[28]; */
+  /* unsigned int  Reserved[25]; */
 };
 
 enum UNRARCALLBACK_MESSAGES {
@@ -206,6 +214,6 @@ int    PASCAL RARGetDllVersion();
 }
 #endif
 
-#pragma pack()
+#pragma pack(pop)
 
 #endif
